#include "parallel.h"
#include "interface.h"
#include <stdio.h>

void parallel_cube_host(int *Command,
                        complex *BRAU, complex *BRAL, complex *STAIRS,
                        short *KTNO, int *LBRI, int *NELI, short *INODIA,
                        complex *RELEM, complex *CELEM,
                        short *KORNO, short *NODIA, int *INDEX,
                        short *IBUSO, complex *VOLT, int *NBUS)
{
   int                      argc = 1;
   char                    *argv[] = {"etmsp31.ht", "\0"};
   static int               Virgin = TRUE;
   static int               old_NBUS = 0;

   ZIP_MAILER              *Mailer;

   static matrix_host_bin   Matrix;
   static vector_host_bin   Vector;

   long                     sec;
   long                     usec;
   int                      i;
   FILE                    *TimingFile;

   printf("cube host -- demo for parallel sparse LU solver\n");

   if (Virgin == TRUE)
   {
      Virgin = FALSE;

      initialize_host(NODE_PGM_NAME, argc, argv, &Matrix, &Vector);
   }

   /* Hosue keeping work */

   Mailer = Matrix.Mailer;

   print("NBUS = %d\n", *NBUS);

   if (old_NBUS != *NBUS)
   {
      int    p;
      int    q;

      int    m;
      int    n;

      g2_pq(Mailer, p, q);

      if (old_NBUS != 0)
      {
         Free(Matrix.Row_Dis -> data);
         Free(Matrix.Row_Dis);
         Free(Matrix.Row_Extra);

         Free(Matrix.Col_Dis -> data);
         Free(Matrix.Col_Dis);
         Free(Matrix.Col_Extra);
         free(Matrix.Row_major_LU_element_Count);
         free(Matrix.Row_major_LU);
      }

      Matrix.M = *NBUS * 2;
      set_distribution(Mailer, Matrix.Row_Dis_Type, CDISTRI_ROW,
                       &(Matrix.Row_Dis),
                       &(Matrix.Row_Distribution),
                       &(Matrix.Row_Inverse_Distribution),
                       &(Matrix.Row_Distribution_limit),
                       &(Matrix.Row_Distribution_init),
                       Matrix.P, p, Matrix.M, &m,
                       (int) Matrix.Row_Extra_param1,
                       (int) Matrix.Row_Extra_param2,
                       &(Matrix.Row_Extra));

      Matrix.N = *NBUS * 2;
      set_distribution(Mailer, Matrix.Col_Dis_Type, CDISTRI_COLUMN,
                       &(Matrix.Col_Dis),
                       &(Matrix.Col_Distribution),
                       &(Matrix.Col_Inverse_Distribution),
                       &(Matrix.Col_Distribution_limit),
                       &(Matrix.Col_Distribution_init),
                       Matrix.Q, q, Matrix.N, &n,
                       (int) Matrix.Col_Extra_param1,
                       (int) Matrix.Col_Extra_param2,
                       &(Matrix.Col_Extra));

      Matrix.NBUS = *NBUS;
      if ((Matrix.Row_major_LU = (element **) malloc(sizeof(element *) * *NBUS))
          == NULL)
      {
         print("parallel_cube_host(): Out of memory!\n");
         exit(1);
      }

      if ((Matrix.Row_major_LU_element_Count = (int *) malloc(sizeof(int)
                                                              * *NBUS)) == NULL)
      {
         print("parallel_cube_host(): Out of memory!\n");
         exit(1);
      }

      for (i = 0; i < *NBUS; i++)
      {
         Matrix.Row_major_LU[i] = NULL;
         Matrix.Row_major_LU_element_Count[i] = 0;
      }

      Vector.M = Matrix.M;
      Vector.N = Matrix.N;

      Vector.Row_Dis = Matrix.Row_Dis;
      Vector.Row_Distribution = Matrix.Row_Distribution;
      Vector.Row_Inverse_Distribution = Matrix.Row_Inverse_Distribution;
      Vector.Row_Distribution_limit = Matrix.Row_Distribution_limit;
      Vector.Row_Extra = Matrix.Row_Extra;

      Vector.Col_Dis = Matrix.Col_Dis;
      Vector.Col_Distribution = Matrix.Col_Distribution;
      Vector.Col_Inverse_Distribution = Matrix.Col_Inverse_Distribution;
      Vector.Col_Distribution_limit = Matrix.Col_Distribution_limit;
      Vector.Col_Extra = Matrix.Col_Extra;

      Vector.NBUS = *NBUS;
      old_NBUS = *NBUS;
   }

   print("Command = %d\n", *Command);
   transmit_command_host(Mailer, *Command);

   /* Command decodeing */
   switch (*Command)
   {
      case DO_LU:
         pack_unfactorized_matrix(&Matrix, BRAU, BRAL, STAIRS, KTNO, LBRI, NELI,
                                  IBUSO, INODIA, *NBUS);
         transmit_matrix_host(Mailer, &Matrix);
         clear_real_matrix_host(&Matrix);

         receive_matrix_host(Mailer, &Matrix);
#if 0
         unpact_factorized_matrix_host(&Matrix, RELEM, CELEM, STAIRS, KORNO,
                                       NODIA, INDEX, IBUSO, INODIA, *NBUS);
#endif
         clear_complex_matrix_host(&Matrix);

         receive_timing_host(Mailer, &sec, &usec);
         print("Timing Data for LU factor: %ld sec %ld usec\n", sec, usec);

         if ((TimingFile = fopen("ETMSP31.timing", "a+")) != NULL)
         {
            fprintf(TimingFile, "Timing Data for LU factor: %ld sec %ld usec\n",
                    sec, usec);
            fclose(TimingFile);
         }
         break;

      case DO_SOLVE:
#if 0
         pack_factorized_matrix_host(&Matrix, RELEM, CELEM, STAIRS, KORNO,
                                     NODIA, INDEX, IBUSO, INODIA, *NBUS);
         transmit_matrix_host(Mailer, &Matrix);
         clear_real_matrix_host(&Matrix);
#endif
         pack_col_vector_host(&Vector, VOLT, IBUSO, *NBUS);
         transmit_col_vector_host(Mailer, &Vector);
         clear_col_vector_host(&Vector);

         receive_row_vector_host(Mailer, &Vector);
         unpack_row_vector_host(&Vector, VOLT, IBUSO, *NBUS);
         clear_row_vector_host(&Vector);

         receive_timing_host(Mailer, &sec, &usec);
         print("Timing Data for forward and back solve: %ld sec %ld usec\n",
               sec, usec);
         if ((TimingFile = fopen("ETMSP31.timing", "a+")) != NULL)
         {
            fprintf(TimingFile,
                   "Timing Data for forward and back solve: %ld sec %ld usec\n",
                    sec, usec);
            fclose(TimingFile);
         }
         break;

      case TERMINATE:
         zip_exit();
#ifdef GHOST
         system("freecube");
#endif
         break;

      default:
         print("parallel_cube_host_(): Unknow command\n");
         exit(1);
   }
}

